//  Copyright (c) CNES  2008
//
//  This software is part of CelestLab, a CNES toolbox for Scilab
//
//  This software is governed by the CeCILL license under French law and
//  abiding by the rules of distribution of free software.  You can  use,
//  modify and/ or redistribute the software under the terms of the CeCILL
//  license as circulated by CEA, CNRS and INRIA at the following URL
//  'http://www.cecill.info'.

package fr.cnes.celestlab.celestlabx.generic;

import java.util.HashMap;
import java.util.Arrays;
import java.util.Map;


/**
 * Class Params
 * This class manages a list of parameters (key/value pairs)
 * The keys are case insensitive
 */
public class Params
{
	// Elements stored in params structure
	class Elem {
		Object[] value; 
		String type; 
		
		Elem(Object[] _value, String _type) {
			value = _value; 
			type = _type; 
		}
	}
	
    private Map<String, Elem> params;

   /**
     * Constructor
     */
    public Params()
    {
        params = new HashMap<String, Elem>();
    }

    /**
     * Add an integer parameter
     * @param name name of parameter {@link String)
     * @param value value of parameter {@link Integer)
     */
    public void addInteger(String name, Integer value)
    {
        Object[] objArray = { value };
        put(name, objArray, "Integer");
    }

    /**
     * Return the value of an integer parameter
     * @param name name of parameter {@link String)
     * @return value of parameter {@link Integer)
     * @throws ParamException
     */
    public Integer getInteger(String name) throws ParamException
    {
        Object object = get1(name, "Integer");
        Integer anInteger = new Integer((Integer) object);
        return anInteger; 
    }

    /**
     * Return the value of an integer array parameter
     * @param name name of parameter {@link String)
     * @return value of parameter {@link Integer) array
     * @throws ParamException
     */
    public Integer[] getIntegerArray(String name) throws ParamException 
    {
        Object[] objectArray = get(name, "Integer");
        Integer[] integerArray = Arrays.copyOf(objectArray, objectArray.length, Integer[].class);
        return integerArray; 
    }

    /**
     * Add a string parameter
     * @param name name of parameter {@link String)
     * @param value value of parameter {@link String)
     */
    public void addString(String name, String value)
    {
        Object[] objArray = { value };
        put(name, objArray, "String");
    }

    /**
     * Add a string array parameter
     * @param name name of parameter {@link String)
     * @param value value of parameter {@link String) array
     */
    public void addStringArray(String name, String[] value)
    {
        Object[] objArray = value;
        put(name, objArray, "String");
    }

    /**
     * Return the value of a string parameter
     * @param name name of parameter {@link String)
     * @return value value of parameter {@link String)
     * @throws ParamException
     */
    public String getString(String name) throws ParamException
    {
        Object object = get1(name, "String");
        String aString = new String((String) object);
        return aString; 
    }

    /**
     * Return the value of a string array parameter
     * @param name name of parameter {@link String)
     * @return value value of parameter {@link String) array
     * @throws ParamException
     */
    public String[] getStringArray(String name) throws ParamException
    {
        Object[] objectArray = get(name, "String");
        String[] aString = Arrays.copyOf(objectArray, objectArray.length, String[].class);
        return aString; 
    }

    /**
     * Add a double parameter
     * @param name name of parameter {@link String)
     * @param value value of parameter {@link Double)

     * @throws ParamException      */
    public void addDouble(String name, Double value)
    {
        Object[] objArray = { value };
        put(name, objArray, "Double");
    }

    /**
     * Return the value of a double parameter
     * @param name name of parameter {@link String)
     * @return value value of parameter {@link Double)
     * @throws ParamException
     */
    public Double getDouble(String name) throws ParamException
    {
        Object object = get1(name, "Double");
        Double aDouble = new Double((Double) object);
        return aDouble; 
    }

    /**
     * Return the value of a double array parameter
     * @param name name of parameter {@link String)
     * @return value value of parameter {@link Double) array
     * @throws ParamException
     */
    public Double[] getDoubleArray(String name) throws ParamException
    {
        Object[] objectArray = get(name, "Double");
        Double[] doubleArray = Arrays.copyOf(objectArray, objectArray.length, Double[].class);
        return doubleArray; 
    }

    /**
     * Add a boolean parameter
     * @param name name of parameter {@link String)
     * @param value value of parameter {@link Boolean)

     * @throws ParamException      */
    public void addBoolean(String name, Boolean value)
    {
        Object[] objArray = { value };
        put(name, objArray, "Boolean");
    }

    /**
     * Return the value of a boolean parameter
     * @param name name of parameter {@link String)
     * @return value value of parameter {@link Boolean)
     * @throws ParamException
     */
    public Boolean getBoolean(String name) throws ParamException
    {
        Object object = get1(name, "Boolean");
        Boolean aBoolean = new Boolean((Boolean) object);
        return aBoolean; 
    }

    /**
     * Return the value of a boolean array parameter
     * @param name name of parameter {@link String)
     * @return value value of parameter {@link Boolean) array
     * @throws ParamException
     */
    public Boolean[] getBooleanArray(String name) throws ParamException
    {
        Object[] objectArray = get(name, "Boolean");
        Boolean booleanArray[] = Arrays.copyOf(objectArray, objectArray.length, Boolean[].class);
        return booleanArray; 
    }

    /**
     * Check parameter exists
     * @param name name of parameter {@link String)
     * @return true if exists {@link Boolean) 
     */
    public boolean exists(String name) 
    {
        if (params.containsKey(name))
            return true;
        else
            return false;
    }

    /**
     * Return the value of a array parameter
     * @param name name of parameter {@link String)
     * @return value of parameter {@link Object) array
     * @throws ParamException
     */
    private void put(String name, Object[] val, String type) 
    {
    	Object[] value = val; 
    	if (value == null) 
    		value = new Object[0]; 
        params.put(name, new Elem(value, type));
    }

    /**
     * Return the value of a array parameter
     * @param name name of parameter {@link String)
     * @return value of parameter {@link Object) array
     * @throws ParamException
     */
    private Object[] get(String name, String type) throws ParamException
    {
    	Elem elem = params.get(name); 
    	
        if (elem != null && type.equals(elem.type))
        	return elem.value; 
        else
            throw new ParamException(name, "not found or wrong type");
    }
    
    /**
     * Return the value of a parameter (1 value)
     * @param name name of parameter {@link String)
     * @return value of parameter {@link Object)
     * @throws ParamException
     */
    private Object get1(String name, String type) throws ParamException
    {
    	Elem elem = params.get(name); 
    	
        if (elem != null && type.equals(elem.type) && elem.value != null && elem.value.length == 1)
        	return elem.value[0]; 
        else
            throw new ParamException(name, "not found or wrong type");
    }

}
