//  Copyright (c) CNES  2008
//
//  This software is part of CelestLab, a CNES toolbox for Scilab
//
//  This software is governed by the CeCILL license under French law and
//  abiding by the rules of distribution of free software.  You can  use,
//  modify and/ or redistribute the software under the terms of the CeCILL
//  license as circulated by CEA, CNRS and INRIA at the following URL
//  'http://www.cecill.info'.

package fr.cnes.celestlab.celestlabx.stela;

import org.apache.log4j.Logger;

import fr.cnes.celestlab.celestlabx.generic.Information;
import fr.cnes.celestlab.celestlabx.generic.Params;
import fr.cnes.los.stela.commons.logic.date.StelaDate;
import fr.cnes.los.stela.commons.model.ephemeris.Bulletin;
import fr.cnes.los.stela.commons.model.ephemeris.FrameTypes;
import fr.cnes.los.stela.commons.model.ephemeris.NatureTypes;
import fr.cnes.los.stela.commons.model.ephemeris.type.Type8PosVel;


/**
 * Class Extrapolation
 * This class provides an interface with Scilab for Stela orbital propagation
 */
public class Extrapolation
{
	private static final Logger logger = Logger.getLogger(Extrapolation.class);

	private Simulation simulation; // simulation context
	private ExtrapolationResults extrapolationResults; // results

	/**
	 * Constructor
	 */
	public Extrapolation()
	{
		simulation = new Simulation();
		extrapolationResults = new ExtrapolationResults();
	}

	/**
	 * This method extrapolates a satellite state vector at given time steps.
	 * @param cjd0: initial epoch, (CJD, TREF time scale)
	 * @param x0: initial orbital elements position ("cireq" type, CelestLab convention, CIRF frame)
	 * @param delta_t: times from cjd0 where the results are computed (s)
	 * @param params: structure containing the model parameters
	 * @param res if 1, the transition matrix is computed (0 otherwise)
	 * @param ut1_tref: UT1 minus TREF (s)
	 * @param tt_tref: TT minus TREF (s)
	 * @return info: status of extrapolation
	 */
	public Information run(double cjd0, double[] x0, double[] delta_t, Params params,
			int res, double ut1_tref, double tt_tref) 
	{

		Information info = _run(cjd0, x0, delta_t, params, res, ut1_tref, tt_tref); 
		simulation.clean();
		return info; 

	}

	// same interface as extrapolate - low level	
	private Information _run(double cjd0, double[] x0, double[] delta_t, Params params,
			int res, double ut1_tref, double tt_tref) 
	{
		logger.debug("Entering extrapolate function");

		// Initialize simulation context
		Information info = simulation.initialize(params, ut1_tref, tt_tref);

		if (info.getStatus() != 0)
			return info;

		if (params == null)
			return new Information(-1, "Parameters object is null");

		// Check we have time steps
		if (delta_t == null || delta_t.length < 1)
			return new Information(-1, "Time steps array is empty or null");

		// Check time steps are positive values
		for (double timeStep: delta_t)
		{
			if (timeStep < 0)
				return new Information(-1, "One or more time steps are negative");
		}

		// Check integration step is strictly positive 
		double integratorStep = simulation.getIntegratorStep(); 

		if (integratorStep <= 0)
			return new Information(-1, "Integration step not positive");

		// Check time steps are multiples of integration steps
		// (required accuracy = integrator step / 1.e10)
		for (int i = 0; i < delta_t.length; i++) {
			long k = Math.round(delta_t[i] / integratorStep); 
			if (Math.abs(delta_t[i] - k * integratorStep) >= integratorStep / 1.e10)
				return new Information(-1, "Time steps not multiple of integration step");
		}

		// Check time steps are increasing
		if (delta_t.length > 1)
		{
			for (int i = 1; i < delta_t.length; i++) {
				if (delta_t[i] <= delta_t[i-1])
					return new Information(-1, "Time steps not increasing");
			}
		}

		// centralEnabled must be true (always included in propagation)
		if (! simulation.getCentralEnabled()) {
			return new Information(-1, "Central body must be taken into account");
		}

		// Set simulation initial state (type8, mean elements, CIRF)
		Bulletin bulletin_init = makeBulletin(cjd0, x0); 
		simulation.setInitialState(bulletin_init);

		// Set simulation duration
		double ndays = delta_t[delta_t.length-1] / 86400;
		simulation.setSimulationDuration(ndays); 

		// Run simulation
		logger.info("Starting extrapolation...");

		boolean computeTransitionMatrix = (res == 1 ? true : false);
		boolean computetimeDerivatives = false;

		int status = 0; 
		String errMsg = ""; 
		double sim_duration = -1; 
		String[] msgXML = simulation.createXML(); // before results ! 

		// Method: 1 = high level, 2 = low level
		// NB: method 2 is experimental (does not work yet) 
		simulation.setSimulationMethod(1); 

		try {
			sim_duration = simulation.simulate(computeTransitionMatrix, computetimeDerivatives); 
			logger.debug(String.format("effective simulation duration: %f (days)", sim_duration / 86400));
			if (sim_duration <= 0) {
				status = -1; 
				errMsg = "Unexpected error - no propagation"; 
			}
		}
		catch (Exception e)
		{
			status = -1;
			if (e.getMessage() != null) errMsg = e.getMessage();
			logger.error(errMsg);
		}

		if (status == 0 && sim_duration < delta_t[delta_t.length-1] - integratorStep/2) {
			logger.info("Premature end of propagation");
			status = 1; 
		}

		logger.info("Extrapolation finished.");

		// initialize results even if errors are detected
		extrapolationResults = new ExtrapolationResults();
		extrapolationResults.computeParamMatrices(simulation, delta_t);

		return new Information(status, errMsg, msgXML);
	}


	/**
	 * Returns the extrapolation results
	 * @return the Extrapolation results
	 */
	public ExtrapolationResults getExtrapolationResults()
	{
		return extrapolationResults;
	}


	/**
	 * Make a "Bulletin" from date and position/velocity in CIRF
	 * @param cjd : date (TREF time scale)
	 * @param x : orbital elements ("cireq" type, CelestLab convention, frame = CIRF)
	 * @param nature: "MEAN" 
	 * @return the "Bulletin"
	 * NB: time scale: converted to UT1 (Stela convention)
	 */
	private Bulletin makeBulletin(double cjd, double[] x)
	{
		double ut1_tref = simulation.getUT1minusTREF(); 
		// Date is set to UT1 time scale
		StelaDate stelaDate = new StelaDate(cjd + ut1_tref / 86400);

		Type8PosVel param = new Type8PosVel(x[0], x[5], x[1], x[2], x[3], x[4],
				NatureTypes.MEAN, FrameTypes.CIRF);

		return new Bulletin(stelaDate, param);
	}

}
