//  Copyright (c) CNES  2008
//
//  This software is part of CelestLab, a CNES toolbox for Scilab
//
//  This software is governed by the CeCILL license under French law and
//  abiding by the rules of distribution of free software.  You can  use,
//  modify and/ or redistribute the software under the terms of the CeCILL
//  license as circulated by CEA, CNRS and INRIA at the following URL
//  'http://www.cecill.info'.

package fr.cnes.celestlab.celestlabx.stela;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;

import fr.cnes.celestlab.celestlabx.generic.Information;
import fr.cnes.celestlab.celestlabx.generic.Params;
import fr.cnes.los.stela.commons.logic.exception.CommonException;
import fr.cnes.los.stela.commons.model.AdvancedParameters;
import fr.cnes.los.stela.commons.model.DefaultValues;
import fr.cnes.los.stela.commons.model.Prop;
import fr.cnes.los.stela.commons.model.ShortPeriodsSettings;
import fr.cnes.los.stela.commons.model.atmosmodel.IAtmosphericModel;
import fr.cnes.los.stela.commons.model.ephemeris.Bulletin;
import fr.cnes.los.stela.commons.model.ephemeris.FrameTypes;
import fr.cnes.los.stela.commons.model.ephemeris.SimulationType;
import fr.cnes.los.stela.commons.model.mathematics.JavaMathAdapter;
import fr.cnes.los.stela.commons.model.spaceobject.SpaceObject;
import fr.cnes.los.stela.commons.model.thirdbody.IThirdBody;
import fr.cnes.los.stela.elib.business.framework.ephemeris.model.EphemerisManager;
import fr.cnes.los.stela.elib.business.framework.simulation.IExtrapolationFeedback;
import fr.cnes.los.stela.elib.business.implementation.atmosmodel.msis00.MSIS00Adapter;
import fr.cnes.los.stela.elib.business.implementation.atmosphericdrag.AtmosphericDragAcc;
import fr.cnes.los.stela.elib.business.implementation.diffeq.DiffEq;
import fr.cnes.los.stela.elib.business.implementation.diffeq.DiffEqTMat;
import fr.cnes.los.stela.elib.business.implementation.diffeq.IForceModel;
import fr.cnes.los.stela.elib.business.implementation.earthpotential.tesseral.TesseralAcc;
import fr.cnes.los.stela.elib.business.implementation.earthpotential.zonal.ZonalAcc;
import fr.cnes.los.stela.elib.business.implementation.earthpotential.zonal.ZonalTMat;
import fr.cnes.los.stela.elib.business.implementation.integrator.Integrator;
import fr.cnes.los.stela.elib.business.implementation.noninertial.ApparentAcc;
import fr.cnes.los.stela.elib.business.implementation.solarradiationpressure.SRPAcc;
import fr.cnes.los.stela.elib.business.implementation.solarradiationpressure.SRPTMat;
import fr.cnes.los.stela.elib.business.implementation.thirdbody.ThirdBodyAcc;
import fr.cnes.los.stela.elib.business.implementation.thirdbody.ThirdBodyTMat;
import fr.cnes.los.stela.elib.business.implementation.thirdbody.moon.model.MoonOrbit;
import fr.cnes.los.stela.elib.business.implementation.thirdbody.sun.model.SunOrbit;
import fr.cnes.los.stela.elib.business.implementation.tide.SolidTidesAcc;
import fr.cnes.los.stela.elib.business.logic.ElibException;
import fr.cnes.los.stela.etoo.business.framework.simulation.SimulationTools;
import fr.cnes.los.stela.etoo.business.implementation.simulation.gtosimulation.GTOSimulation;
import fr.cnes.los.stela.etoo.business.implementation.simulation.gtosimulation.GTOSimulationBuilder;
import fr.cnes.los.stela.processing.flightdynamics.BulletinNatureConverter;
import fr.cnes.los.stela.processing.logic.exception.ProcessingException;


/**
 * Class Simulation
 * This class is used to initialize the simulation context: 
 * - simulation data
 * - (Stela) GTOSimulation object for propagation  
 * - (Stela) GTODiffEq object for time derivatives computation
 * - Some utility functions that are data/model dependent (osc2mean...) 
 * - function to run the propagation
 * 
 * The primary reason for this class is to hide simulation details from other high level classes. 
 * 
 * Note the "reference" time scale in STELA is UT1 (but is TREF in CelestLab interface). 
 * The simulation uses the STELA "GTO" model
 * All orbital elements are "type8" <=> "cireq" in CelestLab. 
 * Integration frame is CIRF. 
 * 
 */
class Simulation
{
	private static final Logger logger = Logger.getLogger(Simulation.class);

	// Simulation data
	private SimulationData data; 

	// STELA GTO simulation object
	private GTOSimulation gtosim; 

	// Thirdbody for mean <-> osculating conversion and for EqDiff creation
	private IThirdBody moon = null; 
	private IThirdBody sun = null; 

	// object (physical properties)
	private SpaceObject spaceObject = null; 

	// Force model
	List<IForceModel> forces; 

	// Ephemeris manager for integration - method 2
	// created by createIntegrator
	EphemerisManager ephManager;

	// type de simulation : 
	// 1: high level, 2: low level
	private int simulationMethod = 0; 



	/**
	 * Constructor
	 * Note: moon, sun, spaceObject: not initialised now. 
	 */
	Simulation()
	{
		data = new SimulationData(); 
		gtosim = (GTOSimulation) new GTOSimulationBuilder().buildSimulation();
		gtosim.setStatisticalAnalysis(false);
	}

	/**
	 * Initializes the simulation context
	 * @param params: structure containing the extrapolation parameters
	 * @param ut1_tref: UT1 minus TREF (s)
	 * @param tt_tref: TT minus TREF (s)
	 * @return info: status of the extrapolation
	 * @throws Exception
	 */
	Information initialize(Params params, double ut1_tref, double tt_tref) 
	{
		int status = 0;
		String errMsg = "";

		try
		{
			// reset internal constants before (possibly) changing default values 
			// NB: solar activity and areo coef : reset to default values 
			// (and will be reloaded) 
			// Important : physical constants not re-initialized  
			DefaultValues.init(); 
			AdvancedParameters.init(); 

			// initialize data structure
			data.initialize(params, ut1_tref, tt_tref); 

			// physical constants file 
			// Warning: if value is changed cannot be reset to default value
			if (!data.physicalConstantsFile.trim().isEmpty()) { 
				Prop.setStelaPhysicalConstants(data.physicalConstantsFile); 
			}

			// solar activity file is reloaded if name is given or if reload wanted
			if (!data.solarActivityFile.trim().isEmpty()) { 
				Prop.setStelaSolarActivity(data.solarActivityFile); 
			}

			// aero coef file is reloaded if name is given or if reload wanted
			if (!data.aeroCoefFile.trim().isEmpty()) { 
				Prop.setStelaDragCoefficient(data.aeroCoefFile); 
			}

			initGTOSimulation(); 
		}
		catch (Exception e)
		{
			status = -1;
			errMsg = e.getMessage();
			logger.error(e.getMessage());
		}

		return new Information(status, errMsg);
	}


	/**
	 * cleaning
	 */
	void clean() 
	{
		data.clean(); 
	}


	/**
	 * returns a eqDiff instance
	 * @param bulletin date + orbital elements
	 */
	DiffEq createEqDiff(Bulletin bulletin) throws Exception
	{       
		DiffEq diffEq = new DiffEq(bulletin, forces);
		return diffEq;
	}


	private DiffEq createDiffEq(Bulletin initialState, boolean computeTransitionMatrix) throws Exception
	{
		ZonalTMat zonalTM = null; 
		ThirdBodyTMat sunTM = null; 
		ThirdBodyTMat moonTM = null; 
		SRPTMat srpTM = null; 
		DiffEq diffEq = null; 

		if (computeTransitionMatrix) {
			if (data.zonal_enabled) {
				zonalTM = new ZonalTMat(JavaMathAdapter.min(data.zonal_maxDeg, AdvancedParameters.TRANSITION_ZONAL_ORDER));
			}
			if (data.thirdBodySun_enabled) {
				sunTM = new ThirdBodyTMat(JavaMathAdapter.min(data.thirdBody_degree,
						AdvancedParameters.TRANSITION_SUN_ORDER), new SunOrbit(data.tt_tref - data.ut1_tref));
			}
			if (data.thirdBodyMoon_enabled) {
				moonTM = new ThirdBodyTMat(JavaMathAdapter.min(data.thirdBody_degree,
						AdvancedParameters.TRANSITION_MOON_ORDER), new MoonOrbit(data.tt_tref - data.ut1_tref));
			}
			if (data.srp_enabled){
				srpTM = new SRPTMat(spaceObject, new SunOrbit(data.tt_tref - data.ut1_tref)); 
			}

			diffEq = new DiffEqTMat(initialState, forces, zonalTM, sunTM, moonTM, srpTM);
		}

		else { 
			diffEq = new DiffEq(initialState, forces);
		}

		return diffEq; 
	}

	/**
	 * returns an integrator instance
	 * @throws CommonException 
	 * @throws ProcessingException 
	 * @throws ElibException 
	 */
	private Integrator createIntegrator(boolean computeTransitionMatrix) throws Exception   
	{    
		// initial state from gtosim object; 

		Bulletin initialState = gtosim.getEphemerisManager().getInitialState(); 
		double simulationDuration = gtosim.getSimulationDuration(); // years
		double integrationStep = gtosim.getIntegratorStep(); // s
		double ephInterval = integrationStep; // ephemeris interval (s) - should be a multiple of the integration step

		double reentryAltitude = gtosim.getReentryAltitude(); // meters
		int paramUpdatePeriod = gtosim.getAtmosRecomputeSteps(); // Drag computing period (in number of integration steps)

		IExtrapolationFeedback feedee = null; // This parameter is used by the progression bar in the GUI

		// Short periods settings : zonal contribution (true/false), zonal order, sun contribution, moon contribution
		ShortPeriodsSettings shortPeriodsSettings = new ShortPeriodsSettings(data.zonal_enabled, data.zonal_maxDeg, 
				data.thirdBodySun_enabled, data.thirdBodyMoon_enabled);

		boolean timeDerivativesFlag = false;

		DiffEq diffEq = createDiffEq(initialState, computeTransitionMatrix); 

		ephManager = new EphemerisManager(initialState);

		double tt_minus_ut1 = data.tt_tref - data.ut1_tref; 
		boolean iterative_mode = false; 

		// Define the integrator and perform the integration
		// (the switch corresponding to the false value is used by the high-level Simulation when iterative mode is activated)
		Integrator integrator = new Integrator("RK6", SimulationType.GTO, ephManager, diffEq, initialState, integrationStep,
				simulationDuration, ephInterval, paramUpdatePeriod, reentryAltitude, feedee, tt_minus_ut1,
				shortPeriodsSettings, timeDerivativesFlag, iterative_mode);

		return integrator; 
	}


	/**
	 * Run simulation - method 1
	 * @return simulation duration (s) or -1
	 * @throws Exception 
	 */
	void setSimulationMethod(int method)  
	{
		simulationMethod = method; 
	}


	/**
	 * Run simulation
	 * @return simulation duration (s) or -1
	 * @throws Exception 
	 */
	double simulate(boolean computeTransitionMatrix, boolean computeTimeDerivatives) throws Exception  
	{
		double sim_duration = 0; 
		boolean fail = false; 

		if (simulationMethod == 1) {
			gtosim.setTransitionMatrixFlag(computeTransitionMatrix);
			gtosim.setTimeDerivativesFlag(computeTimeDerivatives);
			ExtrapolationFeedback feedback = new ExtrapolationFeedback(); 
			fail = gtosim.simulate(feedback); 
		}
		else {
			Integrator integrator = createIntegrator(computeTransitionMatrix); 
			integrator.integrate();
			fail = false; 
		}

		// get simulation duration
		if (!fail) { 
			ArrayList<Bulletin> eph = getEphemerisManager().getEphemerisList();
			double cjd1 = eph.get(0).getDate().getCnesJulianDate();
			double cjd2 = eph.get(eph.size()-1).getDate().getCnesJulianDate();
			sim_duration = (cjd2 - cjd1) * 86400; 
		}
		return sim_duration; 
	}


	/**
	 * Returns the adequate ephemeris manager
	 */
	EphemerisManager getEphemerisManager()
	{
		EphemerisManager eph_manager; 

		if (simulationMethod == 1)
			eph_manager = gtosim.getEphemerisManager();
		else
			eph_manager = ephManager;

		return eph_manager;
	}



	// Set simulation duration
	// duration: days (without margin) 
	// margin included : difference between years of 365.25 days and 
	// years from same calendar dates + additional margin (0.5 days) 
	// because of the special STELA way of counting years
	void setSimulationDuration(double ndays)
	{
		double simulationDuration = ndays + 2.5 + Math.floor(ndays/36525);
		gtosim.setSimulationDuration(simulationDuration / 365.25); 

		logger.debug(String.format("Required simulation duration: %f (days)", ndays));
		logger.debug(String.format("Programmed simulation duration: %f (days)", simulationDuration));
	}

	/**
	 * Sets the initial state for propagation
	 * @param bulletin: orbital state
	 * NB: should be: mean elements, type8, frame = CIRF)
	 */
	void setInitialState(Bulletin bulletin)
	{
		// Initial bulletin
		gtosim.getEphemerisManager().setInitialState(bulletin);
		logger.debug("setInitialState: Bulletin = " + bulletin.toString());
	}

	/**
	 * Gets the integrator step
	 */
	double getIntegratorStep()
	{
		return data.integrator_step;
	}

	/**
	 * Convert to osculating parameters
	 * @return the "converted" Bulletin
	 * @throws Exception 
	 */
	Bulletin mean2osc(Bulletin bulletin_mean) throws Exception 
	{
		Bulletin bulletin = bulletin_mean.clone();  

		ShortPeriodsSettings shortPeriodsSettings = new ShortPeriodsSettings(data.zonal_enabled, data.zonal_maxDeg, 
				data.thirdBodySun_enabled, data.thirdBodyMoon_enabled);

		BulletinNatureConverter.convertToOsculating(bulletin, moon, sun, shortPeriodsSettings); 

		return bulletin; 
	}

	/**
	 * Convert to osculating parameters
	 * @return the "converted" Bulletin
	 * @throws Exception 
	 */
	Bulletin osc2mean(Bulletin bulletin_osc) throws Exception 
	{
		Bulletin bulletin = bulletin_osc.clone();  

		ShortPeriodsSettings shortPeriodsSettings = new ShortPeriodsSettings(data.zonal_enabled, data.zonal_maxDeg, 
				data.thirdBodySun_enabled, data.thirdBodyMoon_enabled);

		BulletinNatureConverter.convertToMean(bulletin, moon, sun, shortPeriodsSettings); 

		return bulletin; 
	}


	/**
	 * @return the value of UT1-TREF from the simulation data
	 */
	double getUT1minusTREF()
	{
		return data.ut1_tref;
	}

	/**
	 * @return the value of "central_enabled" from the simulation data
	 */

	boolean getCentralEnabled()
	{
		return data.central_enabled;
	}

	/**
	 * Initialises the GTO Simulation using the simulation data
	 * Warning: changes some global variables ! 
	 */
	private void initGTOSimulation() 
	{
		spaceObject = new SpaceObject("object", data.mass,
				data.drag_area, data.srp_area, data.srp_coef, data.drag_coef);

		// 
		gtosim.setEphemerisStep(data.integrator_step);
		gtosim.setIntegratorStep(data.integrator_step);

		// time offsets: TT - UT1 = (TT - TREF) - (UT1 - TREF)
		gtosim.setTtMinusUT1(data.tt_tref - data.ut1_tref);

		// Space object
		gtosim.setSpaceObject(spaceObject);

		gtosim.setAtmosModel(data.atmosModelType);

		// Solar activity
		gtosim.setSolarActivity(data.solarActivity);

		// Drag perturbtion
		gtosim.setDragSwitch(data.drag_enabled);
		gtosim.setDragQuadraturePoints(data.drag_nbQuadPoints);
		gtosim.setAtmosRecomputeSteps(data.drag_nbComputeSteps);

		// Solar radiation pressure (SRP)
		gtosim.setSrpSwitch(data.srp_enabled);
		gtosim.setSrpQuadraturePoints(data.srp_nbQuadPoints);

		// Third body
		gtosim.setMoonSwitch(data.thirdBodyMoon_enabled);
		gtosim.setSunSwitch(data.thirdBodySun_enabled);

		// solid tides
		Boolean solidTidesSwich = (data.solidTidesMoon_enabled && data.solidTidesSun_enabled); 
		gtosim.setSolidTidesSwitch(solidTidesSwich); 


		// Zonals terms
		gtosim.setZonalSwitch(data.zonal_enabled); 
		gtosim.setZonalOrder(data.zonal_maxDeg);

		// Tesseral terms
		gtosim.setTesseralSwitch(data.tesseral_enabled);
		gtosim.setTesseralOrder(data.tesseral_maxDeg);

		// filtering duration for tesseral terms => number of steps
		int nbIntegrationStepTesseral = (int) Math.ceil(data.tesseral_minPeriod / data.integrator_step);
		gtosim.setNbIntegrationStepTesseral(nbIntegrationStepTesseral);

		// Reentry altitude
		gtosim.setReentryAltitude(data.reentryAltitude);

		// ----------------------------
		// Initialises Sun and Moon objects using "correct" TT-UT1 value
		// ----------------------------
		double tt_minus_ut1 = data.tt_tref - data.ut1_tref; 

		moon = new MoonOrbit(tt_minus_ut1);
		sun = new SunOrbit(tt_minus_ut1);

		// ----------------------------
		// Creates perturbation objects
		// Used for GTODiffEq creation
		// ----------------------------
		ZonalAcc zonalAcc = null; 
		if (data.zonal_enabled) {
			boolean j22_included = true; 
			zonalAcc = new ZonalAcc(data.zonal_maxDeg, j22_included); 
		}

		TesseralAcc tesseralAcc = null; 
		if (data.tesseral_enabled) {
			tesseralAcc = new TesseralAcc(data.integrator_step, nbIntegrationStepTesseral, data.tesseral_maxDeg);
		}

		ThirdBodyAcc sunAcc = null; 
		if (data.thirdBodySun_enabled) {
			sunAcc = new ThirdBodyAcc(data.thirdBody_degree, sun);
		}

		ThirdBodyAcc moonAcc = null; 
		if (data.thirdBodyMoon_enabled) {
			moonAcc = new ThirdBodyAcc(data.thirdBody_degree, moon);
		}

		AtmosphericDragAcc dragAcc = null; 
		if (data.drag_enabled) {
			// NB: only MSIS2000 is available - no test
			final IAtmosphericModel atmosModel = new MSIS00Adapter();
			dragAcc = new AtmosphericDragAcc(spaceObject, data.drag_nbQuadPoints, 
					atmosModel, data.solarActivity, AdvancedParameters.Z_LIMIT_ATMOS, sun, moon);
		}

		SRPAcc srpAcc = null; 
		if (data.srp_enabled) {
			Boolean eclipses = data.srp_eclipses; // true => consider eclipses
			srpAcc = new SRPAcc(spaceObject, data.srp_nbQuadPoints, eclipses, sun);
		}

		ApparentAcc apparentAcc = null; 
		if (data.appAcc_enabled) {
			FrameTypes integration_frame  = FrameTypes.CIRF; 
			FrameTypes reference_frame  = data.ref_frame; 
			apparentAcc = new ApparentAcc(data.appAcc_nbQuadPoints, integration_frame, 
					reference_frame, data.tt_tref - data.ut1_tref);
		}

		SolidTidesAcc solidTidesAcc = null; 
		if (data.solidTidesSun_enabled || data.solidTidesSun_enabled) {
			// Sun and Moon considered
			solidTidesAcc = new SolidTidesAcc(data.solidTidesSun_enabled, data.solidTidesSun_enabled, sun, moon); 
		}

		forces = new ArrayList<IForceModel>();
		forces.add(zonalAcc);
		forces.add(tesseralAcc);
		forces.add(sunAcc);
		forces.add(moonAcc);
		forces.add(dragAcc);
		forces.add(srpAcc);
		forces.add(apparentAcc);
		forces.add(solidTidesAcc);


		// additional initialisations (global variables)
		// if cannot be set directly

		// DefaultValues.GTO_QUADRATURE_POINTS = data.drag_nbQuadPoints; 
		// DefaultValues.GTO_SRP_QUADRATURE_POINTS = data.srp_nbQuadPoints; 
		AdvancedParameters.REFERENCE_FRAME = data.ref_frame; 
		AdvancedParameters.COMPL_ACC_QUADRATURE_POINTS = data.appAcc_nbQuadPoints; 
		DefaultValues.GTO_ATMOSDRAG_RECOMPUTE_PERIOD_IN_STEPS = data.drag_nbComputeSteps; 
		AdvancedParameters.THIRD_BODY_POT_DEGREE = data.thirdBody_degree; 
		DefaultValues.REENTRY_ALTITUDE_GTO = data.reentryAltitude; 
		AdvancedParameters.COOK_WALL_TEMPERATURE = data.drag_cookWallTemp; 
		AdvancedParameters.COOK_ACCOMODATION = data.drag_cookAccomodCoef; 
		AdvancedParameters.ECLIPSE_SWITCH = data.srp_eclipses; 
		DefaultValues.TT_MINUS_UT1 = data.tt_tref - data.ut1_tref; 

	}


	/**
	 * Returns the simulation parameters XML in STELA XML format
	 * @return the simulation parameters XML content {@link String}
	 */
	String[] createXML()  
	{
		List<String> simXML; 

		try {
			simXML = _createXML();
		} 
		catch (Exception e) {
			simXML = new ArrayList<String>();
		}

		return simXML.toArray(new String[0]);
	}


	/**
	 * Returns the simulation parameters in STELA XML format
	 * @return the simulation parameters XML content {@link String}
	 * @throws Exception
	 */
	private List<String> _createXML() throws Exception
	{
		// Save XML informations to temp file
		File file = File.createTempFile("simulation", ".xml");
		String fileName = file.getAbsolutePath();

		try
		{
			SimulationTools.saveXmlFormat(fileName, gtosim);
		}
		catch(CommonException e)
		{
			logger.error(String.format(
					"Error while saving simulation XML file %s : %s", fileName, e.getMessage()));
			throw e;
		}


		// Load XML
		List<String> simulationXML = new ArrayList<String>();
		BufferedReader bufferedReader = new BufferedReader(new FileReader(fileName));

		try
		{
			String line = bufferedReader.readLine();

			while (line != null)
			{
				simulationXML.add(line);
				line = bufferedReader.readLine();
			}
		}
		catch(Exception e)
		{
			logger.error(String.format("Error while loading temporary simulation XML file %s",
					fileName, e.getMessage()));
			throw e;
		}
		finally
		{
			bufferedReader.close();
		}

		// delete tmp file
		file.delete(); 

		return simulationXML;
	}

}
