//  Copyright (c) CNES  2008
//
//  This software is part of CelestLab, a CNES toolbox for Scilab
//
//  This software is governed by the CeCILL license under French law and
//  abiding by the rules of distribution of free software.  You can  use,
//  modify and/ or redistribute the software under the terms of the CeCILL
//  license as circulated by CEA, CNRS and INRIA at the following URL
//  'http://www.cecill.info'.

package fr.cnes.celestlab.celestlabx.stela;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Arrays;

import org.apache.log4j.Logger;


import fr.cnes.celestlab.celestlabx.generic.ParamException;
import fr.cnes.celestlab.celestlabx.generic.Params;
import fr.cnes.los.stela.commons.model.AdvancedParameters;
import fr.cnes.los.stela.commons.model.DefaultValues;
import fr.cnes.los.stela.commons.model.atmosmodel.AtmosphericModelType;
import fr.cnes.los.stela.commons.model.dragcoeff.DragCoefType;
import fr.cnes.los.stela.commons.model.dragcoeff.IDragCoef;
import fr.cnes.los.stela.commons.model.ephemeris.FrameTypes;
import fr.cnes.los.stela.commons.model.solaractivity.ISolarActivity;
import fr.cnes.los.stela.commons.model.solaractivity.SolarActType;
import fr.cnes.los.stela.elib.business.implementation.dragcoef.constant.model.ConstantDragCoef;
import fr.cnes.los.stela.elib.business.implementation.dragcoef.cook.CookDragCoef;
import fr.cnes.los.stela.elib.business.implementation.dragcoef.variable.model.VariableDragCoef;
import fr.cnes.los.stela.elib.business.implementation.solaractivity.constantmodel.ConstantSolarActivity;
import fr.cnes.los.stela.elib.business.implementation.solaractivity.variablemodel.VariableSolarActivity;


/**
 * Class SimulationData
 * This class is used to initialize the data coming from Scilab
 * (simulation parameters, ut1-tref, tt-tref) 
 * No getter/setter for this class: direct access from the "Simulation" class
 * (for simplification reasons) 
 */
class SimulationData
{
	private static final Logger logger = Logger.getLogger(SimulationData.class);

	// data necessary for a defining a simulation (models, forces, satellite properties)
	double ut1_tref; // UT1 minus TREF (as in CelestLab)
	double tt_tref;  // TT minus TREF  (as in CelestLab)

	// satellite properties
	double mass;
	double drag_area; 
	IDragCoef drag_coef; 
	double srp_area; 
	double srp_coef; 

	AtmosphericModelType atmosModelType; 
	ISolarActivity solarActivity; 
	FrameTypes ref_frame; // reference frame: CIRF or MOD or ICRF

	// perturbations taken into account 
	boolean central_enabled; 
	boolean drag_enabled; 
	boolean srp_enabled; 
	boolean thirdBodyMoon_enabled; 
	boolean thirdBodySun_enabled; 
	boolean zonal_enabled; 
	boolean tesseral_enabled; 
	boolean appAcc_enabled;  
	boolean solidTidesMoon_enabled; 
	boolean solidTidesSun_enabled; 

	// degree / order of Earth potential 
	int zonal_maxDeg; 
	int tesseral_maxDeg; 
	double tesseral_minPeriod; 

	// integrator (mandatory)
	double integrator_step; 

	// advanced parameters 

	int drag_nbQuadPoints; 
	int drag_nbComputeSteps; 
	int srp_nbQuadPoints; 
	boolean srp_eclipses; 
	int appAcc_nbQuadPoints; 

	int thirdBody_degree; 

	double reentryAltitude; 
	double drag_cookWallTemp; 
	double drag_cookAccomodCoef; 

	// data files
	String physicalConstantsFile = ""; 
	String solarActivityFile = ""; 
	String aeroCoefFile = ""; 

	// true if solarActivityFile has been copied
	boolean solarActivityFileDup = false; 

	/**
	 * Constructor
	 * Does nothing
	 */
	SimulationData()
	{
	}

	/**
	 * Initializes the parameters of the simulation
	 * Note the "reference" time scale in STELA is UT1
	 * The simulation uses the STELA "GTO" model
	 * @param params: structure containing the extrapolation parameters
	 * All parameters should be present
	 * @param ut1_tref: UT1 minus TREF (s)
	 * @param tt_tref: TT minus TREF (s)
	 * @return info: status (error message) 
	 * @throws Exception
	 */
	void initialize(Params params, double param_ut1_tref, double param_tt_tref) throws Exception
	{
		// integrator
		integrator_step = params.getDouble("integrator_step");
		
        logger.debug(String.format("%s = %s", "integrator_step", Double.toString(integrator_step)));

		// time
		ut1_tref = param_ut1_tref; 
		tt_tref = param_tt_tref; 
		
        logger.debug(String.format("%s = %s", "ut1_tref", Double.toString(ut1_tref)));
        logger.debug(String.format("%s = %s", "tt_tref", Double.toString(tt_tref)));


		// object's properties
		mass = params.getDouble("mass");
		drag_area = params.getDouble("drag_area");
		drag_coef = paramsDragCoef(params); 
		srp_area = params.getDouble("srp_area"); 
		srp_coef = params.getDouble("srp_coef"); 
		
        logger.debug(String.format("%s = %s", "mass", Double.toString(mass)));
        logger.debug(String.format("%s = %s", "drag_area", Double.toString(drag_area)));
        logger.debug(String.format("%s = %s", "srp_area", Double.toString(srp_area)));
        logger.debug(String.format("%s = %s", "srp_coef", Double.toString(srp_coef)));

		// solar activity 
		solarActivity = paramsSolarActivity(params); 

		central_enabled = params.getBoolean("central_enabled"); 
		drag_enabled = params.getBoolean("drag_enabled"); 
		srp_enabled = params.getBoolean("srp_enabled"); 

        logger.debug(String.format("%s = %s", "central_enabled", Boolean.toString(central_enabled)));
        logger.debug(String.format("%s = %s", "drag_enabled", Boolean.toString(drag_enabled)));
        logger.debug(String.format("%s = %s", "srp_enabled", Boolean.toString(srp_enabled)));


		// 3rd body
		thirdBodySun_enabled = false;
		thirdBodyMoon_enabled = false;
		if (params.getBoolean("thirdBody_enabled")) {
			String[] bodies = paramsBodies(params, "thirdBody_bodies"); 
			thirdBodySun_enabled = Arrays.asList(bodies).contains("Sun");
			thirdBodyMoon_enabled = Arrays.asList(bodies).contains("Moon");
		}		

        logger.debug(String.format("%s = %s", "thirdBodySun_enabled", Boolean.toString(thirdBodySun_enabled)));
        logger.debug(String.format("%s = %s", "thirdBodyMoon_enabled", Boolean.toString(thirdBodyMoon_enabled)));

        
		zonal_enabled = params.getBoolean("zonal_enabled"); 
		tesseral_enabled = params.getBoolean("tesseral_enabled"); 

        logger.debug(String.format("%s = %s", "zonal_enabled", Boolean.toString(zonal_enabled)));
        logger.debug(String.format("%s = %s", "tesseral_enabled", Boolean.toString(tesseral_enabled)));

		// solid tides : not considered by default "solidTides_enabled" not in structure (for compatibility)
		solidTidesSun_enabled = false; 
		solidTidesMoon_enabled = false; 
		if (params.exists("solidTides_enabled") && params.getBoolean("solidTides_enabled")) {
			// bodies: Sun and Moon by default
			String bodies[] = {"Sun", "Moon"}; 
			if (params.exists("solidTides_bodies")) {
				bodies = paramsBodies(params, "solidTides_bodies"); 
			}

			solidTidesSun_enabled = Arrays.asList(bodies).contains("Sun");
			solidTidesMoon_enabled = Arrays.asList(bodies).contains("Moon");

			// temporary limitation : Sun and moon or none
			if ( (solidTidesSun_enabled || solidTidesMoon_enabled) && 
					!(solidTidesSun_enabled && solidTidesMoon_enabled)) {
				throw new ParamException("solidTides_bodies", "Only Sun AND Moon accepted");
			}
		}	
		
        logger.debug(String.format("%s = %s", "solidTidesSun_enabled", Boolean.toString(solidTidesSun_enabled)));
        logger.debug(String.format("%s = %s", "solidTidesMoon_enabled", Boolean.toString(solidTidesMoon_enabled)));


		// apparent acceleration
		appAcc_enabled = true;
		if (params.exists("appAcc_enabled")) {
			appAcc_enabled = params.getBoolean("appAcc_enabled"); 
		}
		// reference frame
		ref_frame = paramsRefFrame(params); 
		if (ref_frame == FrameTypes.CIRF) appAcc_enabled = false; 

		if (!appAcc_enabled) ref_frame = FrameTypes.CIRF; 
		
        logger.debug(String.format("%s = %s", "appAcc_enabled", Boolean.toString(appAcc_enabled)));
        logger.debug(String.format("%s = %s", "ref_frame", ref_frame));


		// number of coefficients
		zonal_maxDeg = params.getInteger("zonal_maxDeg"); 
		tesseral_maxDeg = params.getInteger("tesseral_maxDeg"); 
		tesseral_minPeriod = params.getDouble("tesseral_minPeriod"); 

        logger.debug(String.format("%s = %s", "zonal_maxDeg", Integer.toString(zonal_maxDeg)));
        logger.debug(String.format("%s = %s", "tesseral_maxDeg", Integer.toString(tesseral_maxDeg)));
        logger.debug(String.format("%s = %s", "tesseral_minPeriod", Double.toString(tesseral_minPeriod)));

        
		// advanced parameters

		if (params.exists("drag_atmosphericModel"))
			atmosModelType = paramsDragAtmosphericModel(params); 
		else
			atmosModelType = AtmosphericModelType.MSIS00; 

		if (params.exists("drag_nbQuadPoints"))
			drag_nbQuadPoints = params.getInteger("drag_nbQuadPoints"); 
		else
			drag_nbQuadPoints = DefaultValues.GTO_QUADRATURE_POINTS; 

		logger.debug(String.format("%s = %s", "drag_nbQuadPoints", Integer.toString(drag_nbQuadPoints)));

        
		if (params.exists("srp_nbQuadPoints"))
			srp_nbQuadPoints = params.getInteger("srp_nbQuadPoints"); 
		else
			srp_nbQuadPoints = DefaultValues.GTO_SRP_QUADRATURE_POINTS; 

		logger.debug(String.format("%s = %s", "srp_nbQuadPoints", Integer.toString(srp_nbQuadPoints)));

		
		if (params.exists("srp_eclipses"))
			srp_eclipses = params.getBoolean("srp_eclipses"); 
		else
			srp_eclipses = true; 

		logger.debug(String.format("%s = %s", "srp_eclipses", Boolean.toString(srp_eclipses)));

		
		if (params.exists("appAcc_nbQuadPoints"))
			appAcc_nbQuadPoints = params.getInteger("appAcc_nbQuadPoints"); 
		else
			appAcc_nbQuadPoints = AdvancedParameters.COMPL_ACC_QUADRATURE_POINTS; 

		logger.debug(String.format("%s = %s", "appAcc_nbQuadPoints", Integer.toString(appAcc_nbQuadPoints)));

		
		if (params.exists("drag_nbComputeSteps"))
			drag_nbComputeSteps = params.getInteger("drag_nbComputeSteps"); 
		else
			drag_nbComputeSteps = DefaultValues.GTO_ATMOSDRAG_RECOMPUTE_PERIOD_IN_STEPS; 

		logger.debug(String.format("%s = %s", "drag_nbComputeSteps", Integer.toString(drag_nbComputeSteps)));

		
		if (params.exists("thirdBody_degree"))
			thirdBody_degree = params.getInteger("thirdBody_degree"); 
		else
			thirdBody_degree = AdvancedParameters.THIRD_BODY_POT_DEGREE; 

		logger.debug(String.format("%s = %s", "thirdBody_degree", Integer.toString(thirdBody_degree)));

		
		if (params.exists("reentryAltitude"))
			reentryAltitude = params.getDouble("reentryAltitude"); 
		else
			reentryAltitude = DefaultValues.REENTRY_ALTITUDE_GTO; 

		logger.debug(String.format("%s = %s", "reentryAltitude", Double.toString(reentryAltitude)));

		
		if (params.exists("drag_cookWallTemp"))
			drag_cookWallTemp = params.getDouble("drag_cookWallTemp"); 
		else
			drag_cookWallTemp = AdvancedParameters.COOK_WALL_TEMPERATURE; 

		logger.debug(String.format("%s = %s", "drag_cookWallTemp", Double.toString(drag_cookWallTemp)));

		if (params.exists("drag_cookAccomodCoef"))
			drag_cookAccomodCoef = params.getDouble("drag_cookAccomodCoef"); 
		else
			drag_cookAccomodCoef = AdvancedParameters.COOK_ACCOMODATION; 

		logger.debug(String.format("%s = %s", "drag_cookAccomodCoef", Double.toString(drag_cookAccomodCoef)));

		
		if (params.exists("physicalConstantsFile")) {
			physicalConstantsFile = params.getString("physicalConstantsFile"); 
	        logger.debug(String.format("%s = %s", "physicalConstantsFile", physicalConstantsFile));

	        File f = new File(physicalConstantsFile); 
			if (!f.canRead()) {
				throw new ParamException("physicalConstantsFile", "File not found");
			}
		}

		if (params.exists("solarActivityFile")) {
			solarActivityFile = params.getString("solarActivityFile"); 
	        logger.debug(String.format("%s = %s", "solarActivityFile", solarActivityFile));
	        
			// Copy of file if extension is "act"
			if (drag_enabled && solarActivity.getSolActType() == SolarActType.VARIABLE) { 
				if (extension(solarActivityFile).equals(".act")) {
					String tmp = copySolarActivityFile(solarActivityFile); 
					if (tmp != null) {
						solarActivityFile = tmp; 
						solarActivityFileDup = true; 
					}
					else {
						throw new ParamException("solarActivityFile", "File copy failed");
					}
			        logger.debug(String.format("%s = %s", "solarActivityFile  (effective)", solarActivityFile));
				}
				File f = new File(solarActivityFile); 
				if (!f.canRead()) {
					throw new ParamException("solarActivityFile", "File not found");
				}
			}
		}			

		if (params.exists("aeroCoefFile")) {
			if (drag_enabled && drag_coef.getDragCoefType() == DragCoefType.CONSTANT) { 
				aeroCoefFile = params.getString("aeroCoefFile"); 
				File f = new File(aeroCoefFile); 
				if (!f.canRead()) {
					throw new ParamException("aeroCoefFile", "File not found");
				}
			}
		}

	}


	/**
	 * cleaning
	 */
	void clean() 
	{
		if (solarActivityFileDup) {
			File file = new File(solarActivityFile); 
			file.delete(); 
		}
	}


	// ----------------------------------------------
	// Utilities (internal functions) : 
	// extract information from "params" structure
	// ----------------------------------------------

	/**
	 * Extract solar activity
	 */
	private ISolarActivity paramsSolarActivity(Params params) throws Exception
	{
		ISolarActivity solarActivity;
		String dragSolarActivityType = params.getString("drag_solarActivityType");
        logger.debug(String.format("%s = %s", "drag_solarActivityType", dragSolarActivityType));

		if (dragSolarActivityType.equals("constant")) {
			solarActivity = new ConstantSolarActivity();
			double drag_solarActivityAp = params.getDouble("drag_solarActivityAp"); 
			double drag_solarActivityFlux = params.getDouble("drag_solarActivityFlux"); 

			solarActivity.setConstantAP(drag_solarActivityAp);
			solarActivity.setConstantF107(drag_solarActivityFlux);
	        logger.debug(String.format("%s = %s", "drag_solarActivityAp", Double.toString(drag_solarActivityAp)));
	        logger.debug(String.format("%s = %s", "drag_solarActivityFlux", Double.toString(drag_solarActivityFlux)));

			return solarActivity;
		}
		else if (dragSolarActivityType.equals("variable")) {
			solarActivity = new VariableSolarActivity();
			return solarActivity;
		}
		else throw new ParamException("drag_solarActivityType",
				String.format("Unknown value '%s'", dragSolarActivityType));
	}

	/**
	 * Extract atmospheric model
	 */
	private AtmosphericModelType paramsDragAtmosphericModel(Params params) throws ParamException
	{
		String dragAtmosphericModel = params.getString("drag_atmosphericModel");
        logger.debug(String.format("%s = %s", "dragAtmosphericModel", dragAtmosphericModel));

		if (dragAtmosphericModel.equals("NRLMSISE-00"))
			return AtmosphericModelType.MSIS00;
		else if (dragAtmosphericModel.equals("US76"))
			return AtmosphericModelType.US76;		
		else
			throw new ParamException("drag_atmosphericModel",
					String.format("Unknown value '%s'", dragAtmosphericModel));
	}

	/**
	 * Extract drag coefficient
	 */
	private IDragCoef paramsDragCoef(Params params) throws ParamException
	{
		String dragCoefType = params.getString("drag_coefType");
        logger.debug(String.format("%s = %s", "dragCoefType", dragCoefType));

        if (dragCoefType.equals("variable"))
			return new VariableDragCoef();
		else if (dragCoefType.equals("constant"))
			return new ConstantDragCoef(params.getDouble("drag_coef"));
		else if (dragCoefType.equals("cook"))
			return new CookDragCoef();
		else
			throw new ParamException("drag_coefType",
					String.format("Unknown value: '%s'", dragCoefType));
	}


	/**
	 * Extract list of bodies ("Sun" or "Moon" only accepted) 
	 */
	private String[] paramsBodies(Params params, String name) throws ParamException
	{
		String[] bodies = params.getStringArray(name);
		
		for (String body : bodies) { 
			if (!body.equals("Sun") && !body.equals("Moon")) { 
				throw new ParamException(name,
						String.format("Unknown value: '%s'", body));
			}
		}
		
		StringBuffer buf = new StringBuffer(); 
		for (String body : bodies) { 
			buf.append(body); 
			buf.append(" "); 
		}
        logger.debug(String.format("%s = %s", name, buf));


		return bodies;
	}


	/**
	 * Extract reference frame from parameters
	 */
	private FrameTypes paramsRefFrame(Params params) throws ParamException
	{
		String ref_frame = params.getString("ref_frame"); 
		FrameTypes frame_type = FrameTypes.CIRF; 

		if (ref_frame.equals("CIRS") || ref_frame.equals("CIRF") || ref_frame.equals("ECI")) {
			frame_type = FrameTypes.CIRF; 
		}
		else if (ref_frame.equals("MOD")) {
			frame_type = FrameTypes.CELESTIAL_MEAN_OF_DATE; 
		}
		else if (ref_frame.equals("ICRS") || ref_frame.equals("ICRF") || 
				ref_frame.equals("GCRS") || ref_frame.equals("GCRF")) {
			frame_type = FrameTypes.ICRF; 
		}
		else {
			throw new ParamException("ref_frame",
					String.format("Not allowed frame '%s'", ref_frame));
		}

        return frame_type;
	}


	/**
	 * extension of path ("." included) 
	 * (should go somewhere else...)
	 */
	private String extension(String path)
	{

		String name = path; 
		String suffix = ""; 
		int index; 

		index = name.lastIndexOf("/");
		if (index >= 0) name = path.substring(index+1); 

		index = name.lastIndexOf(".");
		if (index >= 0) suffix = name.substring(index); 

		return suffix; 
	}

	/**
	 * copy of solar activity file ("act" format)
	 * => add STELA header so that STELA accepts the file
	 * => remove lines that starts with "#"  
	 */
	private String copySolarActivityFile(String path) 
	{

		BufferedReader input = null; 
		BufferedWriter output = null; 
		File tmpfile = null; 

		try {
			tmpfile = File.createTempFile("solarActivity", ".tmp"); 
			input = new BufferedReader(new FileReader(path));
			output = new BufferedWriter(new FileWriter(tmpfile));

			output.write("# version 1\n"); 

			String line; 

			while ((line = input.readLine()) != null) {
				if (line.startsWith("#")) continue; 
				output.write(line); 
				output.newLine(); 
			}

		} catch (IOException e1) {
			if (tmpfile != null) tmpfile.delete(); 
			tmpfile = null; 

		} finally {
			// Finalization
			try {
				if (input != null) input.close();
				if (output != null) output.close();
			} catch (IOException e) {
			} 	
		}

		String path_copy = null; 
		if (tmpfile != null) path_copy = tmpfile.getPath(); 

		return path_copy; 
	}


}
